from datetime import datetime
from .location import Location
from . import unit_registry, quantity


class State2:
    def __init__(self, timestamp, datafile):

        self.timestamp = timestamp
        self.vessel = None
        self.symbology = None
        self.latitude = None
        self.longitude = None
        self.heading = None
        self.speed = None
        self.depth = None
        self.text_label = None

        # Initialize pint's unit registry object
        self.unit_registry = unit_registry

    def print(self):
        print(
            "State: Timestamp: {} Vessel: {} Symbology: {} Latitude: {} Longitude: {} Heading: {} Speed: {} Depth: {} TextLabel: {}".format(
                self.timestamp,
                self.vessel,
                self.symbology,
                self.latitude,
                self.longitude,
                self.heading,
                self.speed,
                self.depth,
                self.text_label,
            )
        )

    def set_speed(self, speed: quantity):
        self.speed = speed

    def set_heading(self, heading: quantity):
        self.heading = heading

    def set_location(self, lat_val: float, long_val: float):
        self.location = (lat_val, long_val)

    def set_location_obj(self, location):
        self.location = location

    def set_longitude(self, long_val: float):
        self.longitude = long_val

    def get_timestamp(self):
        return self.timestamp

    def get_platform(self):
        return self.vessel

    def get_symbology(self):
        return self.symbology

    def get_location(self):
        return (self.latitude, self.longitude)

    def get_heading(self):
        return self.heading

    def get_speed(self):
        return self.speed

    def get_depth(self):
        return self.depth

    def get_text_label(self):
        return self.text_label



import os
import unittest

from pepys_import.file.replay_parser import ReplayParser
from pepys_import.file.nmea_parser import NMEAParser
from pepys_import.file.core_parser import CoreParser
from pepys_import.file.file_processor import FileProcessor


FILE_PATH = os.path.dirname(__file__)
BAD_DATA_PATH = os.path.join(FILE_PATH, "sample_data_bad")
DATA_PATH = os.path.join(FILE_PATH, "sample_data")


class SampleParserTests(unittest.TestCase):
    def test_process_folders_not_descending(self):
        processor = FileProcessor("single_level.db")

        processor.register(ReplayParser())
        processor.register(NMEAParser())

        # try bad file
        exception = False
        try:
            processor.process(BAD_DATA_PATH, None, False)
        except Exception:
            exception = True
        self.assertTrue(exception)

        # now good one
        processor.process(DATA_PATH, None, False)

    def test_process_folders_descending(self):
        processor = FileProcessor("descending.db")

        processor.register(ReplayParser())
        processor.register(NMEAParser())

        # try bad file
        exception = False
        try:
            processor.process(BAD_DATA_PATH, None, True)
        except Exception:
            exception = True
        self.assertTrue(exception)

        # now good one
        processor.process(DATA_PATH, None, True)


if __name__ == "__main__":
    unittest.main()



from .core_parser import CoreParser
from pepys_import.core.formats import unit_registry, quantity
from pepys_import.core.formats.state2 import State2
from datetime import datetime


class NMEAParser(CoreParser):
    def __init__(self):
        super().__init__("NMEA File Format")

    def can_accept_suffix(self, suffix):
        return suffix.upper() == ".LOG" or suffix.upper() == ".TXT"

    def can_accept_filename(self, filename):
        return True

    def can_accept_first_line(self, first_line):
        return "$POSL" in first_line

    def can_process_file(self, file_contents):
        return True

    def process(self, data_store, path, file_contents, data_file_id):
        print("NMEA parser working on " + path)

        line_num = 0
        lat_tok = None
        lat_hem_tok = None
        long_tok = None
        long_hem_tok = None
        date_tok = None
        time_tok = None
        hdg_tok = None
        spd_tok = None

        ctr = 0

        for line in file_contents:

            ctr += 1

            if ctr > 5000:
                break

            tokens = line.split(",")

            line_num += 1

            if len(tokens) > 0:

                msg_type = tokens[1]
                if msg_type == "DZA":
                    date_tok = tokens[2]
                    time_tok = tokens[3]
                elif msg_type == "VEL":
                    spd_tok = tokens[6]
                elif msg_type == "HDG":
                    hdg_tok = tokens[2]
                elif msg_type == "POS":
                    lat_tok = tokens[3]
                    lat_hem_tok = tokens[4]
                    long_tok = tokens[5]
                    long_hem_tok = tokens[6]

                # do we have all we need?
                if date_tok and spd_tok and hdg_tok and lat_tok:

                    # and finally store it
                    with data_store.session_scope():
                        datafile = data_store.search_datafile_by_id(data_file_id)
                        platform = data_store.add_to_platforms_from_rep(
                            "Toure", "Ferry", "FR", "Public"
                        )
                        sensor = data_store.add_to_sensors_from_rep(
                            platform.name + "_GPS", platform
                        )

                        date_time = self.parse_timestamp(date_tok, time_tok)
                        state = State2(date_time, sensor)

                        loc = self.parse_location(
                            lat_tok, lat_hem_tok, long_tok, long_hem_tok
                        )

                        state.set_location_obj(loc)
                        state.set_speed(
                            float(spd_tok) * unit_registry.metre / unit_registry.second
                        )
                        state.set_heading(float(hdg_tok) * unit_registry.degree)

                        data_store.add_state_to_states(
                            state, datafile, sensor,
                        )

                        date_tok = None
                        spd_tok = None
                        hdg_tok = None
                        lat_tok = None

    def parse_timestamp(self, date, time):
        if len(date) == 6:
            formatStr = "%y%m%d"
        else:
            formatStr = "%Y%m%d"

        if len(time) == 6:
            formatStr += "%H%M%S"
        else:
            formatStr += "%H%M%S.%f"

        return datetime.strptime(date + time, formatStr)

    def parse_location(self, lat, lat_hem, lon, long_hem):
        latDegs = float(lat[0:2])
        latMins = float(lat[2:4])
        latSecs = float(lat[4:])
        latDegs = latDegs + latMins / 60 + latSecs / 60 / 60

        lonDegs = float(lon[0:3])
        lonMins = float(lon[3:5])
        lonSecs = float(lon[5:])
        lonDegs = lonDegs + lonMins / 60 + lonSecs / 60 / 60

        if lat_hem == "S":
            latDegs = -1 * latDegs

        if lat_hem == "W":
            lonDegs = -1 * lonDegs

        return (latDegs, lonDegs)



from .core_parser import CoreParser
from pepys_import.core.formats.state2 import State2
from datetime import datetime
from pepys_import.core.formats.location import Location
from pepys_import.core.formats import unit_registry, quantity


class ReplayParser(CoreParser):
    def __init__(self):
        super().__init__("Replay File Format")

    def can_accept_suffix(self, suffix):
        return suffix.upper() == ".REP" or suffix.upper() == ".DSF"

    def can_accept_filename(self, filename):
        return True

    def can_accept_first_line(self, first_line):
        return True

    def can_process_file(self, file_contents):
        return True

    def process(self, data_store, path, file_contents, data_file_id):
        print("Rep parser working on " + path)
        line_num = 0
        for line in file_contents:

            line_num += 1

            if line.startswith(";"):
                pass
            else:
                tokens = line.split()

                if len(tokens) < 15:
                    print(
                        "Error on line {} not enough tokens: {}".format(line_num, line)
                    )
                    return False

                # separate token strings
                date_token = tokens[0]
                time_token = tokens[1]
                vessel_name_token = tokens[2]
                symbology_token = tokens[3]
                lat_degrees_token = tokens[4]
                lat_mins_token = tokens[5]
                lat_secs_token = tokens[6]
                lat_hemi_token = tokens[7]
                long_degrees_token = tokens[8]
                long_mins_token = tokens[9]
                long_secs_token = tokens[10]
                long_hemi_token = tokens[11]
                heading_token = tokens[12]
                speed_token = tokens[13]
                depth_token = tokens[14]

                if len(tokens) >= 16:
                    # TODO: join back into single string, or extract full substring
                    self.text_label = tokens[15:]

                if len(date_token) != 6 and len(date_token) != 8:
                    print(
                        "Line {}. Error in Date format {}. Should be either 2 of 4 figure date, followed by month then date".format(
                            line_num, date_token
                        )
                    )
                    return False

                # Times always in Zulu/GMT
                if len(time_token) != 6 and len(time_token) != 10:
                    print(
                        "Line {}. Error in Time format {}. Should be HHMMSS[.SSS]".format(
                            line_num, time_token
                        )
                    )
                    return False

                timestamp = self.parse_timestamp(date_token, time_token)

                # creata state, to store the data
                new_state = State2(timestamp, data_file_id)

                new_state.vessel = vessel_name_token.strip('"')

                symbology_values = symbology_token.split("[")
                if len(symbology_values) >= 1:
                    if len(symbology_values[0]) != 2 and len(symbology_values[0]) != 5:
                        print(
                            "Line {}. Error in Symbology format {}. Should be 2 or 5 chars".format(
                                line_num, symbology_token
                            )
                        )
                        return False
                if len(symbology_values) != 1 and len(symbology_values) != 2:
                    print(
                        "Line {}. Error in Symbology format {}".format(
                            line_num, symbology_token
                        )
                    )
                    return False

                new_state.symbology = symbology_token

                new_state.latitude = self.degrees_for(
                    lat_degrees_token, lat_mins_token, lat_secs_token, lat_hemi_token
                )

                new_state.longitude = self.degrees_for(
                    long_degrees_token,
                    long_mins_token,
                    long_secs_token,
                    long_hemi_token,
                )

                try:
                    valid_heading = float(heading_token)
                except ValueError:
                    print(
                        "Line {}. Error in heading value {}. Couldn't convert to a number".format(
                            line_num, heading_token
                        )
                    )
                    return False
                if 0.0 > valid_heading >= 360.0:
                    print(
                        "Line {}. Error in heading value {}. Should be be between 0 and 359.9 degrees".format(
                            line_num, heading_token
                        )
                    )
                    return False

                # Set heading as degree(quantity-with-unit) object
                new_state.set_heading(valid_heading * unit_registry.degree)

                try:
                    valid_speed = float(speed_token)
                except ValueError:
                    print(
                        "Line {}. Error in speed value {}. Couldn't convert to a number".format(
                            line_num, speed_token
                        )
                    )
                    return False

                # Set speed as knots(quantity-with-unit) object
                new_state.set_speed(valid_speed * unit_registry.knot)

                try:
                    if depth_token == "NaN":
                        self.depth = 0.0
                    else:
                        self.depth = float(depth_token)
                except ValueError:
                    print(
                        "Line {}. Error in depth value {}. Couldn't convert to a number".format(
                            line_num, depth_token
                        )
                    )
                    return False

                # and finally store it
                with data_store.session_scope():
                    datafile = data_store.search_datafile_by_id(data_file_id)
                    platform = data_store.add_to_platforms_from_rep(
                        new_state.get_platform(), "Fisher", "UK", "Public"
                    )
                    sensor = data_store.add_to_sensors_from_rep(
                        platform.name + "_GPS", platform
                    )
                    data_store.add_state_to_states(
                        new_state, datafile, sensor,
                    )

    def degrees_for(self, degs, mins, secs, hemi: str):
        if hemi.upper() == "S" or hemi.upper() == "W":
            factor = -1
        else:
            factor = 1
        return factor * (float(degs) + float(mins) / 60 + float(secs) / 60 / 60)

    def parse_timestamp(self, date, time):
        if len(date) == 6:
            formatStr = "%y%m%d"
        else:
            formatStr = "%Y%m%d"

        if len(time) == 6:
            formatStr += "%H%M%S"
        else:
            formatStr += "%H%M%S.%f"

        return datetime.strptime(date + time, formatStr)



import os

from pepys_import.core.store.data_store import DataStore


class FileProcessor:
    def __init__(self, filename=None):
        self.parsers = []
        if filename == None:
            self.filename = ":memory:"
        else:
            self.filename = filename

    def process(
        self, folder: str, data_store: DataStore = None, descend_tree: bool = True
    ):
        """Process this folder of data
        
        :param folder: Folder path
        :type folder: String
        :param data_store: Database
        :type data_store: DataStore
        :param descend_tree: Whether to recursively descend through the folder tree
        :type descend_tree: bool
        """

        processed_ctr = 0

        # check folder exists
        if not os.path.isdir(folder):
            raise Exception("Folder not found: {}".format(folder))

        # get the data_store
        data_store = DataStore("", "", "", 0, self.filename, db_type="sqlite")
        data_store.initialise()

        # make copy of list of parsers
        good_parsers = self.parsers.copy()

        filename, file_extension = os.path.splitext(folder)

        # capture path in absolute form
        abs_path = os.path.abspath(folder)

        # decide whether to descend tree, or just work on this folder
        if descend_tree:
            # loop through this folder and children
            for current_path, folders, files in os.walk(abs_path):
                for file in files:
                    processed_ctr = self.process_file(
                        file, current_path, good_parsers, data_store, processed_ctr
                    )
        else:
            # loop through this folder
            for file in os.scandir(abs_path):
                if file.is_file():
                    current_path = os.path.join(abs_path, file)
                    processed_ctr = self.process_file(
                        file, current_path, good_parsers, data_store, processed_ctr
                    )

        print("Files got processed:" + str(processed_ctr) + " times")

    def get_first_line(self, path):
        """Retrieve the first line from the file
        
        :param path: Full tile path
        :type path: String
        :return: First line of text
        :rtype: String
        """
        try:
            with open(path, "r", encoding="windows-1252") as f:
                first_line = f.readline()
                return first_line
        except UnicodeDecodeError:
            return None
        finally:
            f.close()

    def get_file_contents(self, full_path: str):
        try:
            with open(full_path, "r", encoding="windows-1252") as f:
                lines = f.read().split("\n")
        finally:
            f.close()
        return lines

    def process_file(self, file, current_path, good_parsers, data_store, processed_ctr):
        filename, file_extension = os.path.splitext(file)
        # make copy of list of parsers
        good_parsers = self.parsers.copy()

        full_path = os.path.join(current_path, file)
        # print("Checking:" + str(full_path))

        # start with file suffxies
        tmp_parsers = good_parsers.copy()
        for parser in tmp_parsers:
            # print("Checking suffix:" + str(parser))
            if not parser.can_accept_suffix(file_extension):
                good_parsers.remove(parser)

        # now the filename
        tmp_parsers = good_parsers.copy()
        for parser in tmp_parsers:
            # print("Checking filename:" + str(parser))
            if not parser.can_accept_filename(filename):
                good_parsers.remove(parser)

        # tests are starting to get expensive. Check
        # we have some file parsers left
        if len(good_parsers) > 0:

            # now the first line
            tmp_parsers = good_parsers.copy()
            first_line = self.get_first_line(full_path)
            for parser in tmp_parsers:
                # print("Checking first_line:" + str(parser))
                if not parser.can_accept_first_line(first_line):
                    good_parsers.remove(parser)

            # get the file contents
            file_contents = self.get_file_contents(full_path)

            # lastly the contents
            tmp_parsers = good_parsers.copy()
            for parser in tmp_parsers:
                if not parser.can_process_file(file_contents):
                    good_parsers.remove(parser)

            # ok, let these parsers handle the file

            with data_store.session_scope():
                data_file = data_store.add_to_datafile_from_rep(
                    filename, file_extension
                )
                data_file_id = data_file.datafile_id

            for parser in good_parsers:
                processed_ctr += 1
                parser.process(data_store, file, file_contents, data_file_id)

        return processed_ctr

    def register(self, parser):
        """Add this parser
        
        :param parser: new parser
        :type parser: CoreParser
        """
        self.parsers.append(parser)



from abc import ABC, abstractmethod


class CoreParser(ABC):
    def __init__(self, name):
        super().__init__()
        self.name = name

    def __str__(self):
        return self.name

    @abstractmethod
    def can_accept_suffix(self, suffix) -> bool:
        """Whether this parser can handle a file with this suffix
        
        :param suffix: File suffix (e.g. ".doc")
        :type suffix: String
        :return: Yes/No
        :rtype: bool
        """
        pass

    @abstractmethod
    def can_accept_filename(self, filename) -> bool:
        """Whether this parser can handle a file with this full filename
        
        :param filename: Full filename
        :type filename: String
        :return: Yes/No
        :rtype: bool
        """
        pass

    @abstractmethod
    def can_accept_first_line(self, first_line) -> bool:
        """Whether this parser can handle a file with this initial line of text
        
        :param first_line: The initial line of text
        :type first_line: String
        :return: Yes/No
        :rtype: bool
        """
        pass

    @abstractmethod
    def can_process_file(self, file_contents) -> bool:
        """Whether this parser can handle this whole file
        
        :param file_contents: Whole file contents
        :type file_contents: String
        :return: Yes/No
        :rtype: bool
        """
        pass

    @abstractmethod
    def process(self, data_store, path, file_contents, data_file):
        """Process this data-file
        
        :param data_store: The data_store
        :type data_store: DataStore
        :param path: File File path
        :type path: String
        :param file_contents: File contents
        :type file_contents: String
        :param data_file: DataFile object
        :type data_file: DataFile
        """
        pass
